
import pytest
import numpy as np
import warnings

from hipose.utils import find_resource_path

warnings.filterwarnings("ignore")


def test_XsensAnalyse_parsing():
    from hipose.data.trial_parsing import extract_xsens_analyse_data

    trial_path = find_resource_path("./resources/sample_trials/ergowear/example_calibration_xsens*")
    data = extract_xsens_analyse_data(trial_path, resample_freq=30.,
                                      reset_position=True, reset_orientation=True)

    assert data["num_samples"] > 500
    for k, v in data.items():
        if isinstance(v, np.ndarray):
            assert len(v) == data["num_samples"]


def test_XsensMtManager_parsing():
    from hipose.data.trial_parsing import extract_mtmanager_data

    trial_path = find_resource_path("./resources/sample_trials/mtwawinda/example_calibration_xsens*")
    data = extract_mtmanager_data(trial_path, ignore_mag=True,
                                  resample_freq=30.,
                                  orient_algorithm="madgwick",
                                  remove_g_vec=True)

    assert data["num_samples"] > 100
    for k, v in data.items():
        if isinstance(v, np.ndarray):
            assert len(v) == data["num_samples"]


def test_XsensMTwAwinda_parsing():
    from hipose.data.trial_parsing import extract_mtwawinda_data

    sensorID_to_segmentName = {
            "00B45C23": "right_hand",
            "00B45C24": "right_foot",
            "00B45C25": "left_shoulder",
            "00B45C26": "left_upper_arm",
            "00B45C27": "left_forearm",
            "00B45C28": "left_hand",
            "00B45C29": "head",
            "00B45C2A": "stern",
            "00B45C2B": "right_upper_leg",
            "00B45C2C": "left_upper_leg",
            "00B45C2D": "pelvis",
            "00B45C2E": "right_lower_leg",
            "00B45C2F": "right_shoulder",
            "00B45C30": "right_forearm",
            "00B45C31": "left_foot",
            "00B45C33": "left_lower_leg",
            "00B45C34": "right_upper_arm",
    }

    trial_path = find_resource_path("./resources/sample_trials/mtwawinda/example_calibration_xsens*")
    data = extract_mtwawinda_data(trial_path,
                                  mapping_sensorid_to_segname=sensorID_to_segmentName,
                                  ignore_mag=True,
                                  resample_freq=30.,
                                  orient_algorithm="madgwick",
                                  remove_g_vec=True)

    assert data["num_samples"] > 100
    for k, v in data.items():
        if isinstance(v, np.ndarray):
            assert len(v) == data["num_samples"]


@pytest.mark.filterwarnings("ignore")
def test_Ergowear_parsing():
    from hipose.data.trial_parsing import extract_ergowear_data

    trial_path = find_resource_path("./resources/sample_trials/ergowear/example_calibration_xsens*")
    data = extract_ergowear_data(trial_path, ignore_mag=True,
                                 resample_freq=30.,
                                 orient_algorithm="madgwick",
                                 remove_g_vec=True,
                                 ref_angles="tpose")

    assert data["num_samples"] > 100
    for k, v in data.items():
        if isinstance(v, np.ndarray):
            assert len(v) == data["num_samples"]


@pytest.mark.filterwarnings("ignore")
@pytest.mark.parametrize("freq", [15., 60.])
@pytest.mark.parametrize("ref_angles", ["npose", "tpose"])
@pytest.mark.parametrize("s2s_calib_method", ["static", "static_mag"])
def test_Ergowear_error(freq, ref_angles, s2s_calib_method):
    from hipose.data.dataset_parsing.parse_cip_ergowear_dataset import map_segs_xsens2ergowear
    from hipose.data.trial_parsing import extract_ergowear_data, extract_xsens_analyse_data
    from hipose.utils import truncate_dict_of_arrays
    from hipose.metrics import qad

    trial_path = find_resource_path("./resources/sample_trials/ergowear/example_calibration_xsens*")
    ergowear_data = extract_ergowear_data(trial_path, ignore_mag=True,
                                          resample_freq=freq,
                                          orient_algorithm="madgwick",
                                          remove_g_vec=False,
                                          ref_angles=ref_angles,
                                          s2s_calib_method=s2s_calib_method,
                                          plot_data=False)

    xsens_data = extract_xsens_analyse_data(trial_path, ref_angles=ref_angles,
                                            resample_freq=freq,
                                            reset_position=True,
                                            reset_orientation=True,
                                            plot_data=False)

    # discard the initial calibration samples (based on ergowear segmentation algorithm)
    ergowear_data = truncate_dict_of_arrays(ergowear_data,
                                            s_idx=ergowear_data["trial_start_idx"])
    xsens_data = truncate_dict_of_arrays(xsens_data,
                                         s_idx=ergowear_data["trial_start_idx"])
    ergowear_data["num_samples"] -= ergowear_data["trial_start_idx"]
    xsens_data["num_samples"] -= ergowear_data["trial_start_idx"]

    qad_counter = []
    for i, (ori_xsens, ori_ergowear) in enumerate(zip(xsens_data['segments_quat'],
                                                       ergowear_data['segments_quat'])):
        # map xsens data to ergowear
        ori_xlikeergo = map_segs_xsens2ergowear(ori_xsens)
        qad_counter.append(qad(ori_ergowear, ori_xlikeergo))

    qad_metric = np.rad2deg(np.mean(qad_counter))
    assert qad_metric < 30         # less than 30º


@pytest.mark.filterwarnings("ignore")
def test_ErgowearDatabase_parsing():
    from hipose.data.dataset_parsing import parse_cip_ergowear_dataset
    from hipose.data.dataset_parsing import report_cip_ergowear_database_status

    dataset_path = find_resource_path("./resources/sample_trials/ergowear/")
    subj_ids = ["example"]
    seq_ids = ["task", "circuit", "sequence", "random", "validation", "calibration"]
    rep_ids = [f"rep{i}" for i in range(1, 10)]
    rep_ids_val = [f"t{i:02d}" for i in range(1, 45)]
    rep_ids_calib = ["xsens"]

    report_cip_ergowear_database_status(dataset_path, verbose=False, save_path=None)
    parse_cip_ergowear_dataset(dataset_path,
                           subject_ids=subj_ids,
                           sequence_ids=seq_ids,
                           repetition_ids=rep_ids + rep_ids_val + rep_ids_calib,
                           resample_freq=30.,
                           save_path=None,
                           ref_angles="tpose",
                           s2s_calib_method="optim",
                           verbose=False,
                           ignore_errors=False,
                           xsens_kwargs=dict(add_g_vec=True),
                           ergowear_kwargs=dict(orient_algorithm="madgwick",
                                                ignore_mag=True,
                                                remove_g_vec=False)
                           )


def test_MTwAwindaDatabase_parsing():
    from hipose.data.dataset_parsing import parse_cip_mtwawinda_dataset
    from hipose.data.dataset_parsing import report_cip_mtwawinda_database_status

    dataset_path = find_resource_path("./resources/sample_trials/mtwawinda/")

    subj_ids = ["example"]
    seq_ids = ["task", "circuit", "sequence", "random", "validation", "calibration"]
    rep_ids = [f"rep{i}" for i in range(1, 10)]
    rep_ids_calib = ["npose", "tpose", "frontpose", "sideways", "xsens"]

    report_cip_mtwawinda_database_status(dataset_path, verbose=False, save_path=None)
    parse_cip_mtwawinda_dataset(dataset_path,
                                subject_ids=subj_ids,
                                sequence_ids=seq_ids,
                                repetition_ids=rep_ids + rep_ids_calib,
                                resample_freq=30.,
                                save_path=None,
                                ref_angles="tpose",
                                s2s_calib_method="optim",
                                verbose=False,
                                ignore_errors=False,
                                xsens_kwargs=dict(add_g_vec=True),
                                mtwawinda_kwargs=dict(orient_algorithm="ekf",
                                                      ignore_mag=False,
                                                      remove_g_vec=False)
                                )


def test_TotalCapture_parsing():
    pass


def test_DIPIMU_parsing():
    pass




import time
import numpy as np


def test_metrics():
    from hipose.metrics import MetricsAnalyser
    from hipose.metrics import (QADistMetric, RelQADistMetric,
                                DistMetric, RelDistMetric, ProcrustesDistMetric,
                                AUCMetric, PCPMetric, TimeMetric)

    from hipose.rotations import quat_random

    # define metrics to evaluate (between ergowear and xsens skeletons)
    metrics_log = MetricsAnalyser(
            metrics=dict(
                    angl_dist=QADistMetric("angl_dist", show_degrees=True),
                    rel_angl_dist=RelQADistMetric("rel_angl_dist", rel_idx=0),
                    auc_angl_dist=AUCMetric("auc_angl_dist", pcp_thresh_range=(0, np.pi), dist="qad"),
                    pcp_angl_dist=PCPMetric("pcp_angl_dist", dist="qad", threshold=np.pi/6),
                    pos_dist=DistMetric("pos_dist", dist="rmse", units="m"),
                    rel_pos_dist=RelDistMetric("rel_pos_dist", dist="mse", rel_idx=0, units="m"),
                    procrustes_pos_dist=ProcrustesDistMetric("procrustes_pos_dist", dist="l2norm", units="m"),
                    auc_pos_dist=AUCMetric("auc_pos_dist", pcp_thresh_range=(0, 0.150), dist="mae"),
                    pcp_pos_dist=PCPMetric("pcp_pos_dist", dist="rmse", threshold=0.5),
                    processing_time=TimeMetric("processing_time", units="ms"),
            )
    )
    metrics_log.reset()

    nseq = 100
    nsegs = 9

    q1 = quat_random(num=nseq*nsegs).reshape(nseq, nsegs, 4)
    q2 = quat_random(num=nseq*nsegs).reshape(nseq, nsegs, 4)

    prange = (-2., 2.)
    p1 = np.random.rand(nseq, nsegs, 3) * (prange[1] - prange[0]) + prange[0]
    p2 = np.random.rand(nseq, nsegs, 3) * (prange[1] - prange[0]) + prange[0]

    # temporal loop
    for t in range(nseq):
        stime = time.perf_counter()

        # compute metrics
        metrics_log.update(
                dict(
                        # angle
                        angl_dist=[q1[t], q2[t]],
                        rel_angl_dist=[q1[t], q2[t]],
                        auc_angl_dist=[q1[t], q2[t]],
                        pcp_angl_dist=[q1[t], q2[t]],
                        # position
                        pos_dist=[p1[t], p2[t]],
                        rel_pos_dist=[p1[t], p2[t]],
                        procrustes_pos_dist=[p1[t], p2[t]],
                        auc_pos_dist=[p1[t], p2[t]],
                        pcp_pos_dist=[p1[t], p2[t]],
                        processing_time=[(time.perf_counter() - stime) * 1000.],
                )
        )

    metrics_log.log_all(save_path=None, show_plots=False, print_metrics=False)
    metrics = metrics_log.get_metrics()
    avgmetrics = metrics_log.get_avg_metrics()




from hipose.utils import find_resource_path


def test_fusion_filter():
    from hipose.data.trial_parsing.extract_xsens_mtmanager import extract_mtmanager_raw_data
    dataset_path = find_resource_path("./resources/sample_trials/mtwawinda/example_calibration_xsens")
    imu_data = extract_mtmanager_raw_data(dataset_path, interpolate_missing=False)

    # initialize filter fusion (example trial has 17 IMUs)
    from hipose.api.fusion_filter import InertialPoseFusionFilter
    ffilts = InertialPoseFusionFilter(
            num_imus=17,
            ignore_mag=False,
            fusion_filter_alg="madgwick",
            s2s_calib_method="static_mag",
            default_data_freq=imu_data["freq"],
    )

    # initialize calibration params from static NPose
    # (example trial has 5s of NPose at the start)
    calib_s = int(imu_data["freq"] * 5)
    ffilts.compute_imus_calibration(acc_calib_data=imu_data["acc"][0:calib_s],
                                    gyr_calib_data=imu_data["gyr"][0:calib_s],
                                    mag_calib_data=imu_data["mag"][0:calib_s])

    # perform filter fusion on trial data to obtain segment orientations
    for idx, (acc, gyr, mag) in enumerate(zip(imu_data["acc"][calib_s:],
                                              imu_data["gyr"][calib_s:],
                                              imu_data["mag"][calib_s:])):
        pred_ori = ffilts.update(acc=acc, gyr=gyr, mag=mag)
        assert pred_ori.shape == (17, 4)



import numpy as np
from scipy.constants.constants import g

from ahrs.filters import AngularRate, Mahony, Madgwick, EKF
from ahrs.common.orientation import acc2q, ecompass

from hipose.rotations import quat_mult, quat_inv, vec_rotate, \
    convert_quat_to_euler, convert_euler_to_quat, quat_avg


def rotate_vectors(*vector_data, rotation, inv=False):
    """Util function to apply the same rotation to multiple vectors.

    Args:
        vector_data([Iterable[np.ndarray[..., 3]]]): one or multiple
            vectors of data to rotate.
        rotation(np.ndarray[..., 4]): quaternion rotation to
            be applied.
        inv(bool): if inverse rotation should be applied.

    Returns:
        (tuple[np.ndarray[..., 3]]): rotated vector data.

    """
    assert len(vector_data) > 0
    rotated_data = []
    for vec in vector_data:
        if len(vec.shape) <= 2:  # scipy can only handle 2 dims
            qrot = (quat_inv(rotation) if inv else rotation)
            rotated_data.append(vec_rotate(vec, qrot))

        else:  # more than 2 dims (flatten extra dims)
            oshape = vec.shape
            vec = vec.reshape(-1, 3)

            # handle only 1 rotation for all data or individual rotation for each sample
            qrot = (rotation
                    if len(rotation.shape) == 1
                    else rotation.reshape(-1, 4))
            qrot = (quat_inv(qrot) if inv else qrot)
            rotated_data.append(vec_rotate(vec, qrot).reshape(*oshape))

    return tuple(rotated_data)


def compute_imu_orientation(acc_data, gyr_data, mag_data=None,
                            algorithm="madgwick", freq=100.0,
                            q0=None, n_init_samples=10):
    """Compute the orientation of IMU sensor using fusion filter.

    Computes the orientation of an imu sensor from the raw inertial
    data using fusion filters. Assumes imu data is in NWU
    reference frame. Output data will also be in NWU reference frame.

    Args:
        acc_data(np.ndarray[Nx3]): imu accelerometer data.
        gyr_data(np.ndarray[Nx3]): imu gyroscope data.
        mag_data(None|np.ndarray[Nx3]): imu magnetometer data. If None,
            algorithms will use versions without magnetometer data
            if possible.
        algorithm(str): algorithm to use to fuse sensor data. Can be
            one of ["integral", "mahony", "madgwick", "ekf"]. Defaults
             to madgwick.
        freq(float): data acquisition frequency.
        q0(None|np.ndarray[4x]): initial orientation quaternion. If
            None then a default orientation is computed using
            the first data sample. q0 should be in NWU referential.
        n_init_samples(int): number of initial samples to initialize
            q0 if its not passed directly as argument.

    Returns:
        (np.ndarray[Nx4]): orientation array of unit quaternion.

    """
    algorithm = algorithm.lower()

    # get initial orientation acc/mag references if not given
    if q0 is not None:
        # convert from NWU to NED from ahrs library
        q0 = quat_mult(convert_euler_to_quat([np.pi, 0, 0], seq="xyz"), q0)
        q0 /= np.linalg.norm(q0, axis=-1, keepdims=True)
    else:
        q0 = (ecompass(np.mean(acc_data[:n_init_samples], axis=0),
                       np.mean(mag_data[:n_init_samples], axis=0),
                       representation="quaternion", frame="NED")
              if mag_data is not None
              else acc2q(np.mean(acc_data[:n_init_samples], axis=0)))
        q0 /= np.linalg.norm(q0, axis=-1, keepdims=True)

    if algorithm == "madgwick":
        orient = Madgwick(acc=acc_data, gyr=gyr_data,
                          mag=mag_data, frequency=freq, q0=q0).Q

    elif algorithm == "mahony":
        orient = Mahony(acc=acc_data, gyr=gyr_data,
                        mag=mag_data, frequency=freq, q0=q0).Q

    elif algorithm == "ekf":
        if mag_data is not None:
            from ahrs.utils.wmm import WMM
            # magnetic info for braga (change accordingly)
            wmm = WMM(latitude=41.5517, longitude=8.4229, height=0.20)
            mref = np.array([wmm.X, wmm.Y, wmm.Z])
        else:
            mref = None
        # (ENU gives expected results for NED frame (bug in ahrs library)?)
        orient = EKF(acc=acc_data, gyr=gyr_data,
                     mag=mag_data, frequency=freq, q0=q0,
                     frame="ENU", magnetic_ref=mref).Q

    elif algorithm == "integral":
        orient = AngularRate(gyr=gyr_data, frequency=freq, q0=q0).Q

    else:
        raise NotImplementedError(
                f"Chosen orientation algorithm not implemented '{algorithm}'! "
                "Choose one of [integral, mahony, madgwick, ekf]")

    if algorithm not in ("integral", ):
        # rotate orientation from NED (computed by library) to NWU (used by us)
        # could probably also rotate raw data before? (but seems to work like this)
        orient = quat_mult(
                convert_euler_to_quat([np.pi, 0, 0], seq="xyz"),
                orient
        )
    return orient


def relative_orientations(orient_data, ref_quat):
    """Transform multiple orientations relative to a reference one.

    Args:
        orient_data(np.ndarray[...xIx4]): orientation quaternion.
        ref_quat(np.ndarray[...1x4]): reference quaternion orientation.

    Returns:
        (np.ndarray[...xIx4]): Array of quaternions with relative
            orientations for each IMU.

    """
    in_shape = np.shape(orient_data)
    assert len(in_shape) > 1
    orient_data = orient_data.reshape((-1, in_shape[-1]))
    return quat_mult(quat_inv(ref_quat), orient_data).reshape(in_shape)


def static_sensor_to_segment_calibration(acc_data, mag_data=None):
    """Compute sensor to segment offset from given data.

    Args:
        acc_data(np.ndarray[Nx3]): imu accelerometer data.
        mag_data(None|np.ndarray[Nx3]): imu magnetometer data. If None,
            algorithm will use versions without magnetometer data.

    Returns:
        (np.ndarray[Nx4]): rotation offsets as quaternions.

    """
    if mag_data is not None:
        # normalize data and compute sensor orientation using TRIAD method
        # (rotation which aligns sensor ref to world ref - gravity / north-pole)
        acc_data /= np.linalg.norm(acc_data, axis=-1, keepdims=True)
        mag_data /= np.linalg.norm(mag_data, axis=-1, keepdims=True)
        s2s_offsets = np.array([ecompass(a, m, frame="NED",
                                         representation="quaternion")
                                for a, m in zip(acc_data, mag_data)])

        # rotate orientation from NED (computed by library) to NWU (used here)
        s2s_offsets = quat_mult(
                convert_euler_to_quat([np.pi, 0, 0], seq="XYZ"),
                s2s_offsets
        )

    else:
        acc_data /= np.linalg.norm(acc_data, axis=-1, keepdims=True)
        s2s_offsets = np.array([acc2q(a) for a in acc_data])

        # rotate orientation from NED (computed by library) to NWU (used by here)
        s2s_offsets = quat_mult(
                convert_euler_to_quat([np.pi, 0, 0], seq="XYZ"),
                s2s_offsets
        )

    s2s_offsets = quat_inv(s2s_offsets)
    return s2s_offsets / np.linalg.norm(s2s_offsets, axis=-1, keepdims=True)


def dynamic_optim_sensor_to_segment_calibration(dyn_calib_traj_acc, dyn_calib_traj_gyr,
                                                ref_traj_acc, ref_traj_gyr,
                                                initial_s2s_guess=None,
                                                n_warmup=5000, n_generations=100,
                                                acc_weight=1., gyr_weight=1.,
                                                smooth_window=201, verbose=False):
    """
    Sensor to segment calibration through CMA-ES optimization.

    Sensor to segment calibration by minimizing the measured error
    between a dynamic trajectory with respect to a reference trajectory,
    based on accelerometer and gyroscope data. Longer trajectories with
    high variability of movement and slow dynamics seem to work best.

    """
    from .metrics import rmse
    from .rotations import quat_random

    try:
        from cmaes import CMA, get_warm_start_mgd
    except ModuleNotFoundError:
        print("cmaes library needs to be installed. Use 'pip install cmaes'. ")

    assert (dyn_calib_traj_acc.shape == dyn_calib_traj_gyr.shape
            == ref_traj_acc.shape == ref_traj_gyr.shape)

    if smooth_window is not None:
        from scipy.signal import savgol_filter
        dyn_calib_traj_acc = savgol_filter(dyn_calib_traj_acc, window_length=smooth_window, polyorder=2, axis=0)
        dyn_calib_traj_gyr = savgol_filter(dyn_calib_traj_gyr, window_length=smooth_window, polyorder=2, axis=0)
        ref_traj_acc = savgol_filter(ref_traj_acc, window_length=smooth_window, polyorder=2, axis=0)
        ref_traj_gyr = savgol_filter(ref_traj_gyr, window_length=smooth_window, polyorder=2, axis=0)

    dyn_calib_traj_acc /= np.linalg.norm(dyn_calib_traj_acc, axis=-1, keepdims=True)
    ref_traj_acc /= np.linalg.norm(ref_traj_acc, axis=-1, keepdims=True)

    # distance function: quaternion angle distance
    def dist_func(s2s_offset):
        align_traj_acc, align_traj_gyr = \
            rotate_vectors(dyn_calib_traj_acc, dyn_calib_traj_gyr,
                           rotation=s2s_offset, inv=True)
        dist_acc = acc_weight * float(rmse(align_traj_acc, ref_traj_acc, reduce=True))
        dist_gyr = gyr_weight * float(rmse(align_traj_gyr, ref_traj_gyr, reduce=True))
        return dist_acc + dist_gyr

    # estimate a promising distribution from random sampling,
    # then generate parameters of the multivariate gaussian distribution.
    init_solutions = ([(initial_s2s_guess, dist_func(initial_s2s_guess))]
                      if initial_s2s_guess is not None
                      else [])

    for q0 in quat_random(num=n_warmup):
        dist = dist_func(q0)
        init_solutions.append((q0, dist))
    ws_mean, ws_sigma, ws_cov = get_warm_start_mgd(init_solutions, gamma=0.1, alpha=0.1)
    ws_mean = ws_mean if initial_s2s_guess is None else initial_s2s_guess.copy()

    # initialize CMAES optimizer
    optim = CMA(mean=ws_mean, sigma=ws_sigma, cov=ws_cov,
                bounds=np.array([[-1, 1], [-1, 1], [-1, 1], [-1, 1]]))

    best_s2s = None
    best_dist = np.inf
    for g in range(n_generations):
        solutions = []
        for _ in range(optim.population_size):
            q = optim.ask()
            dist = dist_func(q)
            solutions.append((q, dist))
        optim.tell(solutions)

        if verbose and (optim.generation - 1) % (n_generations // 10) == 0:
            print(f"gen:{g}   d:{dist:.4f}   q:{np.round(q, 3)}")

        # get one of the solutions (all similar at this point)
        # and save it if its the best so far
        solv_s2s, solv_dist = solutions[0]
        if solv_dist < best_dist:
            best_s2s = solv_s2s.copy()
            best_dist = solv_dist

        if optim.should_stop():
            # (IPOP-CMA-ES) restart CMAES search with population_size * 2
            optim = CMA(mean=ws_mean, sigma=ws_sigma,
                        population_size=optim.population_size * 2)

    if verbose:
        print("\nOptimized s2s values:")
        np.set_printoptions(suppress=True)
        if initial_s2s_guess is not None:
            init_dist = dist_func(initial_s2s_guess)
            print("Init:  d:", init_dist, "euler:", convert_quat_to_euler(initial_s2s_guess, seq="XYZ", degrees=True))
        print("Pred:  d:", best_dist, "euler:", convert_quat_to_euler(best_s2s, seq="XYZ", degrees=True))
        print("------------------------------\n")
    return best_s2s / np.linalg.norm(best_s2s, axis=-1, keepdims=True)


def imus_static_calibration(acc_calib_data, gyr_calib_data, mag_calib_data,
                            manual_align_rots=None, s2s_calib_method=None):
    """Compute calibration parameters from imu data in static pose.

    acc_calib_data(np.ndarray[NxIx3]): raw accelerometer data
        containing n-pose standing calibration samples.
    gyr_calib_data(np.ndarray[NxIx3]): raw gyroscope data containing
        n-pose standing calibration samples.
    mag_calib_data(np.ndarray[NxIx3]): raw magnetometer data containing
        n-pose standing calibration samples.
    manual_align_rots(None|np.ndarray[Ix4]): quaternion rotations to
        apply to manually rotate each sensor to closer to correct
        position.
    s2s_calib_method(None|str): sensor to segment calibration method to
        apply. Can be one of: [None, "manual", "static", "static_mag"].

    Returns:
        (dict[str, np.ndarray]): dictionary containing extracted
            calibration parameters.

    """
    assert s2s_calib_method in [None, "manual", "static", "static_mag"]
    assert s2s_calib_method != "manual" or manual_align_rots is not None

    # compute sensor calibration params (in original ref)
    acc_magn = np.mean(np.linalg.norm(acc_calib_data, axis=-1), axis=0)
    gyr_bias = gyr_calib_data.mean(axis=0)
    mag_magn = np.mean(np.linalg.norm(mag_calib_data, axis=-1), axis=0)

    # compute sensor to segment calibration
    if s2s_calib_method is None:
        s2s_offsets = np.tile([1, 0, 0, 0], reps=(len(gyr_bias), 1))

    elif s2s_calib_method == "manual":
        s2s_offsets = manual_align_rots

    elif s2s_calib_method == "static":
        # add manual calibration to s2s calibration
        # if available, manually rotate IMUs data so they more closely
        # align with desired reference

        if manual_align_rots is not None:
            # need to manually broadcast rotation to match data dimensions
            _rots = np.repeat(
                    manual_align_rots.reshape((1, *manual_align_rots.shape)),
                    len(acc_calib_data),
                    axis=0)

            # rotate imu data to manually aligned ref
            acc_calib_data, gyr_calib_data, mag_calib_data = rotate_vectors(
                    acc_calib_data, gyr_calib_data, mag_calib_data,
                    rotation=_rots, inv=True)

        # calculate s2s static calibration using gravity reference
        s2s_offsets = static_sensor_to_segment_calibration(
                acc_data=np.mean(acc_calib_data, axis=0),
                mag_data=None)

        if manual_align_rots is not None:
            # add manual calibration to s2s calibration
            s2s_offsets = quat_mult(manual_align_rots, s2s_offsets)

    elif s2s_calib_method == "static_mag":
        # calculate s2s calibration using gravity and magnetic field references
        s2s_offsets = static_sensor_to_segment_calibration(
                acc_data=np.mean(acc_calib_data, axis=0),
                mag_data=np.mean(mag_calib_data, axis=0))

    return dict(acc_magn=acc_magn,
                gyr_bias=gyr_bias,
                mag_magn=mag_magn,
                s2s_offset=s2s_offsets)


def apply_imu_calibration(acc_data, gyr_data, mag_data,
                          acc_bias=None, acc_magn=None,
                          gyr_bias=None,
                          mag_hard_iron=None, mag_soft_iron=None, mag_magn=None,
                          s2s_offset=None):
    """
    Applies imu calibration parameters.

    Args:
        acc_data(np.ndarray[Nx3]): imu accelerometer data.
        gyr_data(np.ndarray[Nx3]): imu gyroscope data.
        mag_data(np.ndarray[Nx3]): imu magnetometer data.
        acc_bias(list[float, float, float], None): accelerometer XYZ
            bias params. If None, then no calibration is applied.
        acc_magn(float, None): accelerometer magnitude. Can be applied
            if "acc_bias" is not available.
        gyr_bias(list[float, float, float], None): gyroscope XYZ bias
            params. If None, then no calibration is applied.
        mag_hard_iron(list[float, float, float], None): magnetometer
            XYZ hard-iron bias params. If None, then no calibration
            is applied.
        mag_soft_iron(list[float, float, float], None): magnetometer
            XYZ soft-iron bias params. If None, then no calibration
            is applied.
        mag_magn(float, None): magnetometer magnetic field norm
            (assumes hard/soft bias removed). If None, then no
            calibration is applied.
        s2s_offset(np.ndarray[Nx4]): quaternion rotations to apply to
            IMU data for sensor to segment correction.

    Returns:
        (np.ndarray[Nx3], np.ndarray[Nx3], np.ndarray[Nx3]):
            calibrated_acc, calibrated_gyr, calibrated_mag.

    """
    # apply acc calibration if available
    if acc_bias is not None:
        acc_data = acc_data - acc_bias
    if acc_magn is not None:
        acc_data = (acc_data / acc_magn) * g

    # apply gyr calibration if available
    if gyr_bias is not None:
        gyr_data = gyr_data - gyr_bias

    # apply mag calibration if available
    if mag_hard_iron is not None:
        mag_data = mag_data - mag_hard_iron
    if mag_soft_iron is not None:
        mag_data = mag_data * mag_soft_iron
    if mag_magn is not None:
        mag_data = mag_data / mag_magn

    # apply sensor to segment offset correction
    if s2s_offset is not None:
        acc_data, gyr_data, mag_data = rotate_vectors(
                acc_data, gyr_data, mag_data,
                rotation=s2s_offset, inv=True)
    return acc_data, gyr_data, mag_data


def magnetometer_soft_hard_bias_calibration(mag_calib_data, visualize=False):
    """Computes soft and hard bias parameters from magnetometer data.

    Requires a good sampling over all 3-axis. Follows implementation from:
    https://github.com/kriswiner/MPU6050/wiki/Simple-and-Effective-Magnetometer-Calibration

    Args:
        mag_calib_data(np.ndarray[..., 3]): magnetometer data with
            sensors moving around 360º over all axis.
        visualize(bool): if data before and after calibration should
            be plotted (ex. for debugging).

    Returns:
        (np.ndarray[..., 3], np.ndarray[..., 3]): hard and soft bias
            calibration params.

    """
    from .utils import remove_outliers

    filt_mag_calib_data = remove_outliers(mag_calib_data, std=2.0)
    hxmax, hxmin = filt_mag_calib_data[..., 0].max(), filt_mag_calib_data[..., 0].min()
    hymax, hymin = filt_mag_calib_data[..., 1].max(), filt_mag_calib_data[..., 1].min()
    hzmax, hzmin = filt_mag_calib_data[..., 2].max(), filt_mag_calib_data[..., 2].min()

    # hard iron calibration
    hxb = (hxmax + hxmin) / 2.
    hyb = (hymax + hymin) / 2.
    hzb = (hzmax + hzmin) / 2.
    hard_bias = np.array([hxb, hyb, hzb])

    # simplistic soft iron calibration
    hxs = (hxmax - hxmin) / 2.
    hys = (hymax - hymin) / 2.
    hzs = (hzmax - hzmin) / 2.
    soft_bias = np.array([hxs, hys, hzs])
    soft_bias = soft_bias.mean() / soft_bias

    if visualize:
        import matplotlib.pyplot as plt
        mag_no_bias_data = (filt_mag_calib_data - hard_bias) * soft_bias

        fig, ax = plt.subplots(2, 1, figsize=(7, 10), tight_layout=True)
        for i, (name, data) in enumerate(zip(["Before Calib", "After Calib"],
                                             [mag_calib_data, mag_no_bias_data])):
            ax[i].scatter(data[:, 0], data[:, 1], alpha=0.25)  # x plane
            ax[i].scatter(data[:, 0], data[:, 2], alpha=0.25)  # y plane
            ax[i].scatter(data[:, 1], data[:, 2], alpha=0.25)  # z plane
            ax[i].legend(["x", "y", "z"])
            ax[i].set_title(name)
        plt.show(block=False)
    return hard_bias, soft_bias


def remove_gravity_acceleration(acc, orient, ref_frame="sensor"):
    """
    Compute the free acceleration by removing the gravity component.
    The resulting acceleration can be returned w.r.t to the world
    (linear) or sensor reference (free).

    Args:
        acc(np.ndarray[Nx3]): imu sensor measured acceleration
        orient(np.ndarray[Nx4]): imu orientation quaternion.
        ref_frame(str): referential frame to return acceleration:
            ("sensor" or "world").
    Returns:
        (np.ndarray[Nx3]): acceleration without gravity vector

    """
    assert ref_frame in ["world", "sensor"]
    world_g_vec = np.array([0., 0., -g])
    if ref_frame == "world":
        return vec_rotate(acc, orient) - world_g_vec
    else:
        return acc - vec_rotate(world_g_vec, quat_inv(orient))


def add_gravity_acceleration(acc, orient, ref_frame="sensor"):
    """
    Add gravity vector component to sensor free acceleration. Can deal
    with free (w.r.t to the sensor frame) or linear (w.r.t to the world)
    accelerations. Results are returned in sensor reference frame.

    Args:
        acc(np.ndarray[Nx3]): imu sensor measured acceleration.
        orient(np.ndarray[Nx4]): imu orientation quaternion.
        ref_frame(str): referential frame of the input acceleration:
            ("sensor" or "world").

    Returns:
        (np.ndarray[Nx3]): sensor acceleration with gravity vector

    """
    assert ref_frame in ["world", "sensor"]
    world_g_vec = np.array([0., 0., -g])
    if ref_frame == "world":
        return vec_rotate(acc + world_g_vec, quat_inv(orient))
    else:
        return acc + vec_rotate(world_g_vec, quat_inv(orient))


def magnetometer_heading(mag, acc=None, degrees=False, frame="NWU"):
    """Calculate magnetic north heading from magnetometer data.

    Args:
        mag(np.ndarray[..., 3]): magnetometer data.
        acc(np.ndarray[..., 3]): accelerometer data. Used for
            magnetometer tilt error compensation.
        degrees(bool): if heading should be returned in degrees
            instead of rads.
        frame(str): frame to return results. Can be one of
            ["NED", "NWU"].

    Returns:
        (np.ndarray): compass headings.

    """
    assert frame in ["NED", "NWU"]

    if acc is not None:
        # compensate for magnetometer inclination error
        mag /= np.linalg.norm(mag, axis=-1, keepdims=True)
        acc /= np.linalg.norm(acc, axis=-1, keepdims=True)
        ori = np.array([
                ecompass(a, m, frame="NED", representation="quaternion")
                for a, m in zip(acc.reshape(-1, 3), mag.reshape(-1, 3))
        ])
        heading = convert_quat_to_euler(ori, seq="xyz")[..., -1]
        heading = heading.reshape(*mag.shape[:-1], 1)
    else:
        mag /= np.linalg.norm(mag, axis=-1, keepdims=True)
        heading = np.arctan2(mag[..., 1], mag[..., 0])

    # convert output to desired format
    if frame == "NWU":
        heading = -heading
    if degrees:
        heading = np.rad2deg(heading)
    return heading




import os
import numpy as np
from zipfile import ZipFile

from .rotations import qad as _qad, qcip as _qcip
from .imu import relative_orientations
from .utils import apply_procrustes_alignment

# TODO: add temporal consistency metric (to penalize high movement jitter)


def rmse(pred, target, weights=1.0, reduce=True):
    """
    Computes the RMSE between pred and target.

    Args:
        pred(np.ndarray): array of predicted data.
        target(np.ndarray): array of target data.
        weights(list[float], float): weights for each value.
        reduce(bool): if reduction method should be applied.

    Returns:
        (np.ndarray): rmse between pred and target

    """
    dist = weights * np.sqrt(((pred - target) ** 2).mean(axis=-1))
    return dist.mean() if reduce else dist


def mse(pred, target, weights=1.0, reduce=True):
    """
    Computes the MSE between pred and target.

    Args:
        pred(np.ndarray): array of predicted data.
        target(np.ndarray): array of target data.
        weights(list[float], float): weights for each value.
        reduce(bool): if reduction method should be applied.

    Returns:
        (np.ndarray): mse between pred and target

    """
    dist = weights * ((pred - target) ** 2).mean(axis=-1)
    return dist.mean() if reduce else dist


def mae(pred, target, weights=1.0, reduce=True):
    """
    Computes the MAE between pred and target.

    Args:
        pred(np.ndarray): array of predicted data.
        target(np.ndarray): array of target data.
        weights(list[float], float): weights for each value.
        reduce(bool): if reduction method should be applied.

    Returns:
        (np.ndarray): mae between pred and target

    """
    dist = weights * np.abs(pred - target).mean(axis=-1)
    return dist.mean() if reduce else dist


def l2norm(pred, target, weights=1.0, reduce=True):
    """
    Computes the L2Norm between pred and target.

    Args:
        pred(np.ndarray): array of predicted data.
        target(np.ndarray): array of target data.
        weights(list[float], float): weights for each value.
        reduce(bool): if reduction method should be applied.

    Returns:
        (np.ndarray): mae between pred and target

    """
    dist = weights * np.linalg.norm(pred - target, axis=-1)
    return dist.mean() if reduce else dist


def qad(pred, target, weights=1.0, reduce=True):
    """
    Computes the QAD(quaternion angle distance) between pred and target.

    Args:
        pred(np.ndarray): array of predicted data.
        target(np.ndarray): array of target data.
        weights(list[float], float): weights for each value.
        reduce(bool): if reduction method should be applied.

    Returns:
        (np.ndarray): qad between pred and target

    """
    dist = weights * _qad(pred, target, keepdims=False)
    return dist.mean() if reduce else dist


def pcp(pred, target, dist_func, threshold, reduce=True):
    """
    Computes the percentage of correct predictions (PCP) between pred
    and target.

    Args:
        pred(np.ndarray): array of predicted data.
        target(np.ndarray): array of target data.
        dist_func(function): function to use to compute distance.
        threshold(float): threshold to consider the detection correct.
        reduce(bool): if reduction method should be applied.

    Returns:
        (np.ndarray): pcp between pred and target

    """
    dists = dist_func(pred, target)
    pck = dists < threshold
    return pck.mean() if reduce else pck


def auc(pred, target, dist_func, pck_range, n_values=40):
    """
    Computes percentage of correct (metric) for multiple threshold
    values.

    Args:
        pred(np.ndarray): array of predicted data.
        target(np.ndarray): array of target data.
        dist_func(function): function to use to compute distance.
        pck_range(list[float]): range of data to evaluate.
        n_values(int): number of threshold values to evaluate.

    Returns:
        (np.ndarray): auc between pred and target

    """
    thresholds = np.linspace(1e-12 + pck_range[0], pck_range[1], n_values, dtype=np.float32)
    return [pcp(pred, target, dist_func, threshold, reduce=True)
            for i, threshold in enumerate(thresholds)]


###################################################################################################
###################################################################################################

class Metric(object):
    """
    General interface object to handle data computed using some metric.

    Args:
        name(str): name of the metric.
        mtype(str): type of metric. Available types are ("raw",
            "time", "dist", "pcp", "auc").
        units(str): unit quantity associated with the metric.
        description(str): description of the metric.
        mrange(tuple[float]): metric upper and lower range or None if
            unbounded.
        colnames(list[str]): name of each if metric is computed for
            multiple entries.
        colweights(list[float]): weighting to apply for each entry when
            averaging values along all entries.
        err_thresh(float): desired error threshold for this metric.

    """

    def __init__(self, name, mtype, units="units", description=None, mrange=None,
                 colnames=None, colweights=None, err_thresh=None):
        self.name = name
        self.mtype = mtype
        self.units = units
        self.description = description if description is not None else name
        self.mrange = mrange
        self.colnames = colnames
        self.colweights = colweights if colweights is not None else 1.0
        self.err_thresh = err_thresh
        assert mtype in ("raw", "time", "dist", "pcp", "auc")
        assert (mrange is None or len(mrange) == 2)
        # store data
        self._data = []

    def __len__(self):
        return len(self._data)

    def reset(self):
        """
        Resets computed metrics buffer.
        """
        self._data = []

    @property
    def data(self):
        """
        Returns computed metric data as a np.array.
        """
        return np.array(self._data)

    def update(self, *args):
        """
        Computes and stores metric data from new sample of data. This
        function should be implemented by child classes.
        """
        raise NotImplementedError

###################################################################################################
# raw data
class RawData(Metric):
    """
    Stores raw data, but uses same interface as the Metric class in
    order to access MetricsAnalyser functionality (namely exporting
    data + time plots).
    """
    def __init__(self, name, **kwargs):
        super(RawData, self).__init__(name=name, mtype="raw", **kwargs)

    def update(self, raw_data):
        self._data.append(raw_data)


###################################################################################################
class TimeMetric(Metric):
    """
    Stores metrics related with time (e.g. inference time, etc...)
    """
    def __init__(self, name, units="s", **kwargs):
        super(TimeMetric, self).__init__(name=name, mtype="time", units=units, **kwargs)

    def update(self, t):
        self._data.append(t)


###################################################################################################
# distance metrics
class DistMetric(Metric):
    """
    Stores a distance metric data between predicted and target inputs.
    """
    def __init__(self, name, dist, **kwargs):
        assert dist in ["rmse", "mse", "mae", "l2norm", "qad"]
        self._dist_func = dict(rmse=rmse, mse=mse, mae=mae, l2norm=l2norm, qad=qad)[dist]
        super(DistMetric, self).__init__(name=name, mtype="dist", **kwargs)

    def update(self, pred, target):
        self._data.append(self._dist_func(pred, target, weights=self.colweights, reduce=False))


class RelDistMetric(Metric):
    """
    Stores a distance metric data between predicted and target inputs
    relative to one of the inputs (ex. root_joint).
    """
    def __init__(self, name, dist, rel_idx=0, **kwargs):
        assert dist in ["rmse", "mse", "mae", "l2norm"]
        self._dist_func = dict(rmse=rmse, mse=mse, mae=mae, l2norm=l2norm)[dist]
        self.rel_idx = rel_idx
        super(RelDistMetric, self).__init__(name=name, mtype="dist", **kwargs)

    def update(self, pred, target):
        pred = pred - pred[self.rel_idx]
        target = target - target[self.rel_idx]
        self._data.append(self._dist_func(pred, target, weights=self.colweights, reduce=False))


class ProcrustesDistMetric(Metric):
    """
    Stores distance metric data between predicted and target inputs
    after Procrustes alignment.
    """
    def __init__(self, name, dist, **kwargs):
        assert dist in ["rmse", "mse", "mae", "l2norm"]
        self._dist_func = dict(rmse=rmse, mse=mse, mae=mae, l2norm=l2norm)[dist]
        super(ProcrustesDistMetric, self).__init__(name=name, mtype="dist", **kwargs)

    def update(self, pred, target):
        pred = apply_procrustes_alignment(pred, target)
        self._data.append(self._dist_func(pred, target, weights=self.colweights, reduce=False))


class QADistMetric(Metric):
    """
    Stores quaternion angle distance (QAD) error between predicted
    and target inputs.
    """
    def __init__(self, name, show_degrees=False, **kwargs):
        units = "deg" if show_degrees else "rads"
        mrange = (0., np.pi)
        if show_degrees and "err_thresh" in kwargs:
            kwargs["err_thresh"] = np.rad2deg(kwargs["err_thresh"])
            mrange = (0., 180.)

        self._show_degrees = show_degrees
        super(QADistMetric, self).__init__(name=name, mtype="dist", units=units,
                                           mrange=mrange, **kwargs)

    def update(self, pred, target):
        assert (np.all(((pred >= -1.0) & (pred <= 1.0)))
                and np.all(((target >= -1.0) & (target <= 1.0)))), \
            "QADistMetric inputs need to be normalized quaternions!"
        if self._show_degrees:
            self._data.append(np.rad2deg(qad(pred, target, reduce=False)))
        else:
            self._data.append(qad(pred, target, reduce=False))


class RelQADistMetric(Metric):
    """
    Stores quaternion angle distance (QAD) error between predicted
    and target inputs, with normalized orientations relative to one of
    the inputs (ex. root_joint).
    """
    def __init__(self, name, rel_idx=0, show_degrees=False, **kwargs):
        units = "deg" if show_degrees else "rads"
        mrange = (0., np.pi)
        if show_degrees and "err_thresh" in kwargs:
            kwargs["err_thresh"] = np.rad2deg(kwargs["err_thresh"])
            mrange = (0., 180.)

        self.rel_idx = rel_idx
        self._show_degrees = show_degrees
        super(RelQADistMetric, self).__init__(name=name, mtype="dist", units=units,
                                              mrange=mrange, **kwargs)

    def update(self, pred, target):
        assert (np.all(((pred >= -1.0) & (pred <= 1.0)))
                and np.all(((target >= -1.0) & (target <= 1.0)))), \
            "QADistMetric inputs need to be normalized quaternions!"

        pred = relative_orientations(pred, ref_quat=pred[..., self.rel_idx, :])
        target = relative_orientations(target, ref_quat=target[..., self.rel_idx, :])
        if self._show_degrees:
            self._data.append(np.rad2deg(qad(pred, target, reduce=False)))
        else:
            self._data.append(qad(pred, target, reduce=False))


###################################################################################################
# percentage of correct predictions metrics
class PCPMetric(Metric):
    """
    Percentage of Correct Predictions (PCP). Stores percentage of
    samples error, between predicted and target inputs, bellow the
    defined threshold.
    """
    def __init__(self, name, dist, threshold, units="%", **kwargs):
        assert dist in ["rmse", "mse", "mae", "l2norm", "qad"]
        self._dist_func = dict(rmse=rmse, mse=mse, mae=mae, l2norm=l2norm, qad=qad)[dist]
        super(PCPMetric, self).__init__(name=name, mtype="pcp", mrange=(0.0, 100.0),
                                        units=units, err_thresh=threshold, **kwargs)

    def update(self, pred, target):
        self._data.append(pcp(pred, target, dist_func=self._dist_func,
                              threshold=self.err_thresh, reduce=False))


class AUCMetric(Metric):
    """
    Stores the percentage of correct predictions for multiple
    threshold points.
    """
    def __init__(self, name, dist, pcp_thresh_range, n_thresholds=40, units="%", **kwargs):
        super(AUCMetric, self).__init__(name=name, mtype="auc", mrange=(0.0, 100.0),
                                        units=units, colnames=None, **kwargs)
        assert dist in ["rmse", "mse", "mae", "l2norm", "qad"]
        assert len(pcp_thresh_range) == 2
        self.pck_thresh_range = pcp_thresh_range
        self.n_thresholds = n_thresholds
        self._dist_func = dict(rmse=rmse, mse=mse, mae=mae, l2norm=l2norm, qad=qad)[dist]

    def update(self, pred, target):
        self._data.append(auc(pred, target, dist_func=self._dist_func,
                              pck_range=self.pck_thresh_range,
                              n_values=self.n_thresholds))

###################################################################################################
###################################################################################################


class MetricsAnalyser(object):
    """
    Computes, stores and analyses relevant metrics for the evaluation
    of pose estimation algorithms.

    Args:
        metrics(dict[str, Metric]): dictionary with metrics to compute.
        exp_name(str): name of the experiment. Useful for logging.

    TODO:
        - supports options to plot/export raw data
    """

    def __init__(self, metrics, exp_name="MetricsAnalyser"):
        self.exp_name = exp_name
        self._metrics = metrics
        self._n = 0

    def reset(self):
        self._n = 0
        self._time = []
        for m in self._metrics.values():
            m.reset()

    def update(self, data_dict):
        """
        Updates metrics with new samples of data.

        Args:
            data_dict (dict[str, any]): dictionary containing data for
                each of the initialized metrics. Keys need be the same
                as the names for each metric given on init, values are
                data to be computed.

        """
        # update number of samples + inference time
        self._n += 1
        # update metrics
        for k, v in data_dict.items():
            self._metrics[k].update(*v)

    def get_metrics(self):
        """
        Get values computed for each metric.

        Returns:
            (dict[str, np.ndarray]): dict with computed metric
                values for each metric.

        """
        d = dict()
        for m in self._metrics.values():
            d[m.name] = m.data
        return d

    def get_avg_metrics(self):
        """
        Get average values (and confidence interval) obtained for
        each of the computed metrics.

        Returns:
            (dict[str, list[float, float]]): dict with average (and ci)
                values for each metric.

        """
        d = dict()
        for m in self._metrics.values():
            vals = m.data
            mean = np.nanmean(vals)
            ci = np.nanstd(vals)  # np.mean((st.sem(vals) * st.t.ppf((1 + 0.95) / 2, len(vals) - 1)))
            d[m.name] = [mean, ci]
        return d

    def print_avg_metrics(self, save_path=None, print_metrics=True,
                          extra_info=""):
        """
        Prints averages of all the computed metrics.

        Args:
            save_path(str, None): path to save the plots, or None if
                saving is not wanted.
            print_metrics(bool): if metrics should be printed.
            extra_info(str): extra information to add to the
                terminal log.

        """
        s = f"Experiment:  {self.exp_name}"
        s += f"\nAveraged metrics results for ({self._n}) samples:"
        for m in self._metrics.values():
            vals = m.data
            mean = np.round(np.nanmean(vals), 3)
            ci = np.round(np.nanstd(vals), 3)  # np.mean((st.sem(vals) * st.t.ppf((1 + 0.95) / 2, len(vals) - 1)))
            s += f"\n  {(m.name + ':').ljust(25)} {str(mean)} ± {str(ci)} {m.units}"

        if extra_info:
            s += "\n\nExtraInfo:\n" + extra_info

        if print_metrics:
            print("##########################################")
            print(s)
            print("##########################################")

        if save_path is not None:
            os.makedirs(os.path.dirname(save_path), exist_ok=True)
            with open(save_path, "w") as f:
                f.write(s)
        return s

    def plot_over_time(self, save_path=None, show_plots=True, remove_outliers=None):
        """
        Plots the metrics over time.

        Args:
            save_path(str, None): path to save the plots, or None if
                saving is not wanted.
            show_plots (bool): if plots should be visualized.
            remove_outliers(bool): if extreme outliers should be
                removed (q > percentile ± 3.0 * IQR).

        """

        import pandas as pd
        import matplotlib.pyplot as plt
        import seaborn as sns
        sns.set()

        valid_metrics = [m for m in self._metrics.values()
                         if (m.mtype in ("raw", "time", "dist")) and (len(m) > 0)]

        if not valid_metrics:
            print("No metrics available for temporal error plots!")
            return

        fig, ax = plt.subplots(len(valid_metrics), 1,
                               figsize=(35, 10 * len(valid_metrics)))
        fig.suptitle(self.exp_name, fontsize=54)
        fig.subplots_adjust(left=0.05, right=0.975, bottom=0.05, top=0.925, hspace=0.3)

        if not hasattr(ax, "__iter__"): ax = [ax]
        for i, m in enumerate(valid_metrics):
            # create pandas dataframe to handle data for seaborn
            data_table = pd.DataFrame(data=m.data, columns=m.colnames)

            # remove extreme outliers(3 * IQR)
            if remove_outliers:
                Q1 = data_table.quantile(0.25)
                Q3 = data_table.quantile(0.75)
                IQR = Q3 - Q1
                data_table = data_table[~((data_table < (Q1 - 3.0 * IQR))
                                          | (data_table > (Q3 + 3.0 * IQR))
                                          ).any(axis=1)]

            if data_table.shape[1] > 1: # if multiple columns of data
                # plot metric for each column over time
                sns.lineplot(data=data_table, ax=ax[i], dashes=False,
                             palette=sns.color_palette("nipy_spectral", data_table.shape[1]))

                # plot mean metric over time
                sns.lineplot(data=data_table.mean(axis=1), ax=ax[i],
                             color="dimgray", linewidth=6, dashes=False)
            else:
                sns.lineplot(data=data_table, ax=ax[i])

            # plots metadata
            ax[i].set_title(str(m.name), fontsize=44)
            ax[i].set_xlabel("Step", fontsize=30)
            ax[i].set_ylabel(m.units, fontsize=30)
            ax[i].tick_params(axis="both", which="major", labelsize=22)
            ax[i].legend(prop={"size": 20})
            if m.mrange is not None:
                ax[i].set_ylim(m.mrange[0], m.mrange[1])

            # add error threshold line
            if m.err_thresh is not None:
                ax[i].axhline(y=m.err_thresh, color="k", linestyle="dashed", lw=3)

        if save_path is not None:
            os.makedirs(os.path.dirname(save_path), exist_ok=True)
            fig.savefig(save_path)
        if show_plots:
            plt.show(block=False)

    def plot_boxplot(self, save_path=None, show_plots=True, remove_outliers=True):
        """
        Plots boxplots to analyse error for each column of data.

        Args:
            save_path(str, None): path to save the plots, or None if
                saving is not wanted.
            show_plots (bool): if plots should be visualized.
            remove_outliers(bool): if extreme outliers should be
                removed (q > percentile ± 3.0 * IQR).

        """
        import pandas as pd
        import matplotlib.pyplot as plt
        import seaborn as sns
        sns.set()

        # select valid metrics to show in boxplots (rmse)
        valid_metrics = [m for m in self._metrics.values()
                         if (m.mtype == "dist") and (len(m) > 0)]

        if not valid_metrics:
            print("No rmse metrics available for boxplots!")
            return

        fig, ax = plt.subplots(len(valid_metrics), 1,
                               figsize=(23, 15 * len(valid_metrics)))
        fig.suptitle(self.exp_name, fontsize=54)
        fig.subplots_adjust(left=0.075, right=0.95, bottom=0.15, top=0.92, hspace=0.35)

        if not hasattr(ax, "__iter__"): ax = [ax]   # if only 1 plot
        for i, m in enumerate(valid_metrics):
            # create pandas dataframe to handle data for seaborn
            data_table = pd.DataFrame(data=m.data,
                                      columns=m.colnames)

            # remove extreme outliers(3.0 * IQR)
            if remove_outliers:
                Q1 = data_table.quantile(0.25)
                Q3 = data_table.quantile(0.75)
                IQR = Q3 - Q1
                data_table = data_table[~((data_table < (Q1 - 3.0 * IQR))
                                          | (data_table > (Q3 + 3.0 * IQR))
                                          ).any(axis=1)]

            # add boxplots
            sns.boxplot(data=data_table, ax=ax[i],
                        palette=sns.color_palette("nipy_spectral", data_table.shape[1]))

            # plots metadata
            ax[i].set_title(str(m.name), fontsize=46)
            ax[i].set_xlabel(m.description, fontsize=30)
            ax[i].set_ylabel(m.units, fontsize=30)
            ax[i].set_xticklabels(ax[i].get_xticklabels(), rotation=75)
            ax[i].tick_params(axis="both", which="major", labelsize=22)
            if m.mrange is not None:
                ax[i].set_ylim(m.mrange[0], m.mrange[1])

            # add error threshold line
            if m.err_thresh is not None:
                ax[i].axhline(y=m.err_thresh, color="k", linestyle="dashed", lw=3)

        if save_path is not None:
            os.makedirs(os.path.dirname(save_path), exist_ok=True)
            fig.savefig(save_path)
        if show_plots:
            plt.show(block=False)

    def plot_auc(self, save_path=None, show_plots=True, remove_outliers=True):
        """
        Plots auc plots to analyse PCK across different thresholds.

        Args:
            save_path(str, None): path to save the plots, or None if
                saving is not wanted.
            show_plots (bool): if plots should be visualized.
            remove_outliers(bool): if extreme outliers should be
                removed (q > percentile ± 3.0 * IQR).

        """
        import pandas as pd
        import matplotlib.pyplot as plt
        import seaborn as sns
        sns.set()

        valid_metrics = [m for m in self._metrics.values()
                         if (m.mtype == "auc") and (len(m) > 0)]

        if not valid_metrics:
            print("No AUC metrics available for AUC plots!")
            return

        fig, ax = plt.subplots(len(valid_metrics), 1,
                               figsize=(20, 10 * len(valid_metrics)))
        fig.suptitle(self.exp_name, fontsize=44)
        fig.subplots_adjust(left=0.05, right=0.975, bottom=0.075, top=0.92, hspace=0.3)

        if not hasattr(ax, "__iter__"): ax = [ax]
        for i, m in enumerate(valid_metrics):
            # create pandas dataframe to handle data for seaborn
            m_vals = m.data * 100.0
            thresholds = np.linspace(m.pck_thresh_range[0], m.pck_thresh_range[1],
                                     m_vals.shape[1], dtype=np.float32)
            data_table = pd.DataFrame(data=np.mean(m_vals, axis=0),
                                      index=list(map(str, np.round(thresholds, 4))))

            # remove extreme outliers(3.0 * IQR)
            if remove_outliers:
                Q1 = data_table.quantile(0.25)
                Q3 = data_table.quantile(0.75)
                IQR = Q3 - Q1
                data_table = data_table[~((data_table < (Q1 - 3.0 * IQR))
                                          | (data_table > (Q3 + 3.0 * IQR))
                                          ).any(axis=1)]

            sns.lineplot(data=data_table, ax=ax[i], legend=None, markers=True)

            # plots metadata
            ax[i].set_title(str(m.name).upper(), fontsize=28)
            ax[i].set_xlabel(f"Error Thresholds ({m.description})", fontsize=22)
            ax[i].set_ylabel(m.units, fontsize=22)
            ax[i].set_xticklabels(list(map(str, np.round(thresholds, 3))), rotation=45)
            if m.mrange is not None:
                ax[i].set_ylim(m.mrange[0], m.mrange[1])

            # add error threshold line
            if m.err_thresh is not None:
                ax[i].axvline(x=(m.err_thresh * (len(ax[i].get_xticklabels())
                                                 / (m.pck_thresh_range[1]))) - 0.5,
                              color="k", linestyle="dashed", lw=2)

        if save_path is not None:
            os.makedirs(os.path.dirname(save_path), exist_ok=True)
            fig.savefig(save_path)
        if show_plots:
            plt.show(block=False)

    def export_data(self, save_path):
        """
        Exports all computed metrics to csv and zips them.

        Args:
            save_path(str): file path to save the plots.

        """
        import pandas as pd

        if not self._metrics:
            print("No metrics available to export!")
            return

        os.makedirs(os.path.dirname(save_path), exist_ok=True)
        save_path_name = os.path.dirname(save_path)
        save_file_name = os.path.basename(save_path)

        # export computed metrics to csv
        file_names = []
        for m in self._metrics.values():
            # create dataframe
            m_vals = m.data
            if len(m_vals) > 0:
                data = pd.DataFrame(data=m_vals, columns=m.colnames)

                # export metrics
                fname = save_path_name + f"/{m.name}.csv"
                file_names.append(fname)
                data.to_csv(fname, float_format="%.5f")

        # zip all data
        with ZipFile(save_path_name + "/" + save_file_name.replace(".zip", "") + ".zip", "w") as z:
            for m in file_names:
                z.write(m, os.path.basename(m))
                os.remove(m)

    def log_all(self, save_path=None, show_plots=True, print_metrics=True,
                remove_outliers=True, extra_info="", export=False):
        """
        Utility function to log all options available.

        Args:
            save_path(str, None): path to save the plots, or None if
                saving is not wanted.
            show_plots (bool): if plots should be visualized.
            print_metrics(bool): if metrics should be printed.
            remove_outliers(bool): if extreme outliers should be
                removed (q > percentile ± 3.0 * IQR).
            extra_info(str): extra information to add to the terminal
                output.
            export(bool): if computed metrics should be exported to csv.

        """
        if export:
            self.export_data(save_path + "/metric_data.zip" if save_path is not None else "./")

        self.print_avg_metrics(save_path + "/metrics.txt" if save_path is not None else None,
                               print_metrics=print_metrics, extra_info=extra_info)

        self.plot_boxplot(save_path + "/boxplots.png" if save_path is not None else None,
                          show_plots=show_plots, remove_outliers=remove_outliers)

        self.plot_auc(save_path + "/auc_plots.png" if save_path is not None else None,
                      show_plots=show_plots, remove_outliers=remove_outliers)

        self.plot_over_time(save_path + "/error_plot.png" if save_path is not None else None,
                            show_plots=show_plots, remove_outliers=remove_outliers)



import numpy as np
import pytest

from hipose.skeleton import Skeleton, SkeletonXsens, \
        SkeletonErgowear, SkeletonMTwAwinda, SkeletonXsensUpper

from hipose.rotations import quat_random

# flake8: noqa: F841

def test_skeleton_objects():

    s1 = Skeleton(root_joint=0, joint_names=["j1", "j2", "j3"],
                  segment_starts=[0, 1], segment_ends=[1, 2])
    s1 = SkeletonXsens()
    s2 = SkeletonMTwAwinda()
    s3 = SkeletonErgowear()
    s4 = SkeletonXsensUpper()


def test_skeleton_mappings():
    # TODO: add
    pass


def test_skeleton_kinematics():
    # TODO: add
    pass


@pytest.mark.filterwarnings("ignore")
@pytest.mark.parametrize("skeleton", [SkeletonXsens(),
                                      SkeletonXsensUpper(),
                                      SkeletonMTwAwinda(),
                                      SkeletonErgowear()])
def test_skeleton_joint_angles(skeleton):
    segs_ori = skeleton.remove_dummy_segments(quat_random(skeleton.num_joints))
    jnt_angles = skeleton.segment_orients_to_relative_joint_angles(
            segs_ori, euler_seq="XYZ", degrees=True
    )
    assert len(jnt_angles) == len(segs_ori)
    assert list(jnt_angles[0]) == [0, 0, 0]


def test_skeleton_vizualization():
    # TODO: add
    pass



import pytest
import warnings

warnings.filterwarnings("ignore")
# flake8: noqa: F401


@pytest.mark.filterwarnings("ignore")
def test_requirement_imports():
    import numpy
    import scipy
    import ahrs


@pytest.mark.filterwarnings("ignore")
def test_pkg_imports():
    import hipose

    import hipose.data.trial_parsing
    import hipose.data.trial_parsing.extract_ergowear
    import hipose.data.trial_parsing.extract_xsens_mtmanager
    import hipose.data.trial_parsing.extract_xsens_analyse

    import hipose.data.dataset_parsing
    import hipose.data.dataset_parsing.parse_cip_ergowear_dataset
    import hipose.data.dataset_parsing.parse_cip_mtwawinda_dataset


    import hipose.imu
    import hipose.rotations
    import hipose.metrics
    import hipose.utils




from os import path
import itertools

from setuptools import setup

from hipose import __version__

here = path.abspath(path.dirname(__file__))

requires_list = []
with open(path.join(here, "requirements.txt"), encoding="utf-8") as f:
    for line in f:
        requires_list.append(str(line))

extras = {
        "parse": ["pandas", "openpyxl"],
        "plot":  ["matplotlib", "seaborn"],
        "vis3d": ["PyQt5", "PyOpenGL", "pyqtgraph"],
        "doc":   ["sphinx", "sphinx-rtd-theme", "sphinx-autoapi"],
        "test":  ["pytest", "pytest-flake8", "pytest-cov", "pytest-clarity", "pytest-randomly", "pytest-pydocstyle"]
}
extras["all"] = list(itertools.chain(*extras.values()))

setup(name="hipose",
      version=__version__,
      description="Human whole-body pose estimation using MARG multi-sensor data.",
      author="ManuelPalermo",
      author_email="macpalermo@gmail.com",
      url="https://github.com/ManuelPalermo/HumanInertialPose",
      license='MIT',
      packages=["hipose"],
      install_requires=requires_list,
      extras_require=extras,
      classifiers=[
              "Programming Language :: Python :: 3",
              "License :: OSI Approved :: MIT License",
              "Operating System :: OS Independent"],
      )




import os
import glob

import numpy as np

from scipy import interpolate
from scipy.spatial.transform import Rotation as R, RotationSpline

from copy import deepcopy
from collections import defaultdict

# nested dict replacement
from .rotations import convert_quat_wxyz_to_xyzw, \
    convert_quat_xyzw_to_wxyz, quat_mult, convert_quat_to_euler,\
    convert_euler_to_quat, quat_inv, vec_rotate


nested_dict = lambda: defaultdict(nested_dict)

def convert_nestedddict_to_regular(d):
    """
    Converts nested defaultdict object to regular python nested dicts.
    """
    if isinstance(d, defaultdict):
        d = {k: convert_nestedddict_to_regular(v) for k, v in d.items()}
    return d


def truncate_dict_of_arrays(dict_array, s_idx=0, e_idx=999999999999999,
                            inplace=False):
    """Truncate arrays inside a dictionary to desired start/end idx."""
    if not inplace:
        dict_array = deepcopy(dict_array)

    for k, v in dict_array.items():
        if isinstance(v, (np.ndarray, list)):
            dict_array[k] = v[s_idx:e_idx]
    return dict_array


def select_idx_dict_of_arrays(dict_array, axis_idx_dict,
                              inplace=False, ignore_idx_errors=True):
    """Selects indexes in desired axis in arrays inside a dictionary.

    Args:
        dict_array(dict[str,np.ndarray): dictionary with
            arrays or tensors.
        axis_idx_dict(dict[int,list[int]]): dictionary with indexes to
            select for each axis as {axis:indexes}.
        inplace(bool): if changes should be done on input arrays.
        ignore_idx_errors(bool): if errors on arrays with missing
            dimensions should be ignored.

    Return:
        dict[str,np.ndarray]:

    """
    out_dict_array = (dict_array if inplace else dict())
    for k, v in dict_array.items():
        ix = [axis_idx_dict.get(dim, slice(None)) for dim in range(v.ndim)]
        if isinstance(v, (np.ndarray)):
            try:
                out_dict_array[k] = v[tuple(ix)]
            except IndexError as e:
                if not ignore_idx_errors:
                    raise e
                else:
                    out_dict_array[k] = v
    return out_dict_array


def find_resource_path(path, max_up=5):
    """Recursively looks for files on the current directory and parents.

    Args:
        path(str): base path or files name of resource to search.
        max_up(int): max parent directory from which to recurse.

    Returns:
        (str): path to found resource or None if it was not found.

    """
    for i in range(max_up):
        pmatch = glob.glob(os.path.join("**/", path), recursive=True)
        if not pmatch:
            path = "../" + path
        else:
            path = pmatch[0]
            return path
    return None


def resample_data_frequency_factor(data, factor, axis=0, method="cubic"):
    """
    Resample the data by the desired factor (assumes uniform sampling).

    Args:
        data(np.ndarray): data array to resample.
        factor(float): factor to resample the data. if factor<1.0, decimation
            is performed,  otherwise, interpolation is performed.
        axis(int): index of the axis to interpolate data along.
        method(str): method to use for resampling the data. Defaults to
            cubic spline. When resampling quaternion rotations,
            use "slerp".

    Returns:
        (np.ndarray): The resampled data array.

    """
    x = np.arange(data.shape[axis])
    x_new = np.linspace(0, data.shape[axis] - 1, round(data.shape[axis] * factor), endpoint=True)

    return resample_data_frequency(data, orig_t=x, target_t=x_new,
                                   axis=axis, method=method)


def resample_data_frequency(data, orig_t, target_t, axis=0, method="cubic"):
    """
    Resample the data from original sampling to target.

    Args:
        data(np.ndarray): data array to resample.
        orig_t(np.ndarray): original timestamps for each data point.
        target_t(np.ndarray): target timestamps for resampled points.
        axis(int): index of the axis to interpolate data along.
        method(str): method to use for resampling the data. Defaults to
            cubic spline. When resampling quaternion rotations,
            use "slerp".

    Returns:
        (np.ndarray): The resampled data array.

    """
    if method == "slerp":
        assert axis == 0, "Spherical Rotation Spline only works when axis=0"
        data = convert_quat_wxyz_to_xyzw(data)  # convert quats to scalar_last for scipy
        if len(np.shape(data)) == 3:  # multiple segments
            sampled_data = np.array(
                    [RotationSpline(orig_t, R.from_quat(ori))(target_t).as_quat()
                     for ori in data.transpose((1, 0, 2))
                     ]).transpose((1, 0, 2))
        else:
            sampled_data = RotationSpline(orig_t, R.from_quat(data))(target_t).as_quat()
        return convert_quat_xyzw_to_wxyz(sampled_data)  # convert quats back to scalar_first
    else:
        return interpolate.interp1d(orig_t, data, kind=method, axis=axis,
                                    bounds_error=False, fill_value="extrapolate")(target_t)


def find_low_variability_sections(data, threshold, window_size=21,
                                  axis=0, thresh_method="max"):
    """
    Find sections of data which contain low variability.

    Args:
        data(np.ndarray): data to search.
        window_size(int): window size to average data.
        threshold(float): threshold value to consider low_variability
        axis(int): axis on which to search.
        thresh_method(str): channel reduction method to compare with
            threshold. One of ["mean", "min", "max"]

    Returns:
        (list[int]): indexes of samples with low variability over
            the desired axis

    """
    from scipy.signal import convolve
    from scipy.signal.windows import gaussian

    reduce_dims = list(np.arange(len(data.shape)))
    reduce_dims.remove(axis)
    reduce_dims = tuple(reduce_dims)

    # apply gaussian smoothing over the axis
    kernel = np.expand_dims(gaussian(window_size, std=window_size / 8), axis=reduce_dims)
    smooth_signal = convolve(data, kernel / kernel.sum(), mode="same")

    # calculate ||pointwise derivatives||
    diff = np.abs(np.diff(smooth_signal, axis=0))
    diff[-window_size:window_size, ...] = 0

    # take (mean, min, max) variability over all channels
    if thresh_method == "mean":
        avg_diff = np.mean(diff, axis=reduce_dims)
    elif thresh_method == "min":
        avg_diff = np.min(diff, axis=reduce_dims)
    elif thresh_method == "max":
        avg_diff = np.max(diff, axis=reduce_dims)
    else:
        raise NotImplementedError

    return list(np.where(avg_diff < threshold)[0])


def remove_outliers(data, std=2.0):
    """Statistical outlier removal for each axis independently.

    Args:
        data(np.ndarray): data from which to remove outliers.
        std(float): stdev for vaues to be considered outliers.

    Returns:
        (np.ndarray): data with outliers removed

    """
    assert len(data.shape) == 2
    num, dim = data.shape
    inliers = list(range(num))
    outliers = []
    for ax in range(dim):
        diff = np.append(0., np.diff(data[..., ax]))
        ax_outliers = np.where(np.abs(diff) > np.abs(np.mean(diff)) + (std * np.std(diff)))[0]
        outliers.extend(ax_outliers)

    inliers = np.array(list(set(inliers) - set(outliers)))
    return data[inliers]


def reset_skeleton_position(pos_args_list, pos_ref, axis2reset=(True, True, False)):
    """
    Resets position of data by removing a reference position.

    Args:
        pos_args_list(list[np.ndarray]): list with arrays of 3d points
            from which to remove the reference.
        pos_ref (np.ndarray[3x]): reference position.
        axis2reset (tuple[bool]): operation mask. Only axis (xyz) with
            True will be reset.

    Returns:
        (list[np.ndarray]): 3d arrays with reset position.

    """
    return [(p - pos_ref * axis2reset) for p in pos_args_list]


def reset_skeleton_orientation(rot_ref, orient_arg_list=(),
                               pos_args_list=(), vec_args_list=(),
                               axis2reset=(False, False, True)):
    """
    Reset orientation by removing a reference rotation.

    Args:
        rot_ref(np.ndarray): reference rotation.
        orient_arg_list(tuple[np.ndarray]): list with arrays of
            quaternion orientations.
        pos_args_list(tuple[np.ndarray]): list with arrays of 3d
            positions around a center point (assumed to be
            index 0 - root).
        vec_args_list(tuple[np.ndarray]): list with arrays of 3d
            vectors.
        axis2reset(tuple[bool, bool, bool]): operation mask. Only axis
            (xyz) with True will be reset.

    Returns:
        (list[np.ndarray], list[np.ndarray], list[np.ndarray]):
            orientation and position arrays with reset orientations.

    """
    # reset heading (rot over z-axis, so that subject is facing forward)
    inv_init_ori = quat_inv(
            convert_euler_to_quat(
                    convert_quat_to_euler(rot_ref, seq="xyz") * axis2reset,
                    seq="xyz",
            )
    )

    reset_ori_data = []
    for ori in orient_arg_list:
        ori_shape = ori.shape
        ori = quat_mult(inv_init_ori, ori.reshape(-1, 4)).reshape(*ori_shape)
        reset_ori_data.append(ori)

    reset_pos_data = []
    for pos in pos_args_list:
        pos_shape = pos.shape

        # center rotation to origin before rotating
        init_pos = pos[0, 0]
        pos = pos - init_pos

        # apply rotation
        pos = vec_rotate(pos.reshape(-1, 3), inv_init_ori).reshape(*pos_shape)

        # restore position to original
        pos = pos + init_pos
        reset_pos_data.append(pos)

    reset_vec_data = []
    for vec in vec_args_list:
        vec_shape = vec.shape
        vec = vec_rotate(vec.reshape(-1, 3), inv_init_ori).reshape(*vec_shape)
        reset_vec_data.append(vec)

    return reset_ori_data, reset_pos_data, reset_vec_data


def apply_procrustes_alignment(pred, target):
    """
    Applies procrustes alignment to find closest fit from "pred" to "target" data.

    Args:
        pred(np.ndarray[Nx3]): source array to be aligned.
        target(np.ndarray[Nx3]): target array for alignment.

    Returns:
        predicted_aligned - Procrustes aligned data

    """
    pred = pred[np.newaxis, ...]
    target = target[np.newaxis, ...]

    muX = np.mean(target, axis=1, keepdims=True)
    muY = np.mean(pred, axis=1, keepdims=True)

    X0 = target - muX
    Y0 = pred - muY

    normX = np.sqrt(np.sum(X0 ** 2, axis=(1, 2), keepdims=True))
    normY = np.sqrt(np.sum(Y0 ** 2, axis=(1, 2), keepdims=True))

    X0 /= normX
    Y0 /= normY

    H = np.matmul(X0.transpose(0, 2, 1), Y0)
    U, s, Vt = np.linalg.svd(H)
    V = Vt.transpose(0, 2, 1)
    R = np.matmul(V, U.transpose(0, 2, 1))

    # Avoid improper rotations (reflections), i.e. rotations with det(R) = -1
    sign_detR = np.sign(np.expand_dims(np.linalg.det(R), axis=1))
    V[:, :, -1] *= sign_detR
    s[:, -1] *= sign_detR.flatten()
    R = np.matmul(V, U.transpose(0, 2, 1))  # Rotation

    tr = np.expand_dims(np.sum(s, axis=1, keepdims=True), axis=2)

    a = tr * normX / normY  # Scale
    t = muX - a * np.matmul(muY, R)  # Translation

    # Perform rigid transformation on the input
    predicted_aligned = a * np.matmul(pred, R) + t
    return np.squeeze(predicted_aligned)




import numpy as np
from scipy.spatial.transform import Rotation as R
from scipy.spatial.transform.rotation import Slerp


'''
TODO: extend all quat operations to work on NDimensional arrays 
    -(perform operations over the last axis)
    -can use (inpt)flaten->op->reshape(out) sequence of operations
'''


def convert_quat_xyzw_to_wxyz(quat):
    """Convert quaternion from xyzw to wxyz representation."""
    return quat[..., [3, 0, 1, 2]]


def convert_quat_wxyz_to_xyzw(quat):
    """Convert quaternion from wxyz to xyzw representation."""
    return quat[..., [1, 2, 3, 0]]


def quat_diff(q1, q2, quat_scalar_last=False):
    """Compute the rotation of quaternions arrays q1 relative to q2.

    Args:
        q1(np.ndarray[Nx4]): quaternion arrays.
        q2(np.ndarray[Nx4]): quaternion arrays.
        quat_scalar_last(bool): quaternion convention to use.

    Returns:
        (np.ndarray[Nx4]): relative quaternion rotation arrays.

    """
    if quat_scalar_last:
        return (R.from_quat(q2).inv() * R.from_quat(q1)).as_quat()
    else:
        q1 = convert_quat_wxyz_to_xyzw(q1)
        q2 = convert_quat_wxyz_to_xyzw(q2)
        q = (R.from_quat(q2).inv() * R.from_quat(q1)).as_quat()
        return convert_quat_xyzw_to_wxyz(q)


def quat_mult(q1, q2, quat_scalar_last=False):
    """Compute rotation resulting from the rotation of quat1 by quat2.

    Args:
        q1(np.ndarray[Nx4]):  quaternion arrays.
        q2(np.ndarray[Nx4]):  quaternion arrays.
        quat_scalar_last(bool): quaternion convention to use.

    Returns:
        (np.ndarray[Nx4]): resulting rotation quaternion arrays.

    """
    if quat_scalar_last:
        return (R.from_quat(q1) * R.from_quat(q2)).as_quat()
    else:
        q1 = convert_quat_wxyz_to_xyzw(q1)
        q2 = convert_quat_wxyz_to_xyzw(q2)
        q = (R.from_quat(q1) * R.from_quat(q2)).as_quat()
        return convert_quat_xyzw_to_wxyz(q)


def quat_inv(quat, quat_scalar_last=False):
    """Inverts rotation represented by quaternion."""
    if quat_scalar_last:
        return R.from_quat(quat).inv().as_quat()
    else:
        quat = convert_quat_wxyz_to_xyzw(quat)
        q = R.from_quat(quat).inv().as_quat()
        return convert_quat_xyzw_to_wxyz(q)


def quat_avg(quat, quat_scalar_last=False):
    """Averages the rotations represented by quaternions."""
    if quat_scalar_last:
        return R.from_quat(quat).mean()
    else:
        quat = convert_quat_wxyz_to_xyzw(quat)
        q = R.from_quat(quat).mean().as_quat()
        return convert_quat_xyzw_to_wxyz(q)


def quat_random(num=None, quat_scalar_last=False):
    """Creates random quaternion rotations."""
    if quat_scalar_last:
        return R.random(num=num).as_quat()
    else:
        q = R.random(num=num).as_quat()
        return convert_quat_xyzw_to_wxyz(q)


def vec_rotate(vec, quat, quat_scalar_last=False):
    """Rotates a vector by the rotation represented by a quaternion."""
    if quat_scalar_last:
        return R.from_quat(quat).apply(vec)
    else:
        quat = convert_quat_wxyz_to_xyzw(quat)
        return R.from_quat(quat).apply(vec)


def quat_slerp(q1, q2, orig_points, target_points, axis=0, quat_scalar_last=False):
    if quat_scalar_last:
        orig_quats = R.from_quat(np.concatenate([q1, q2], axis=axis))
        slerp = Slerp(orig_points, orig_quats)
        return slerp(target_points).as_quat()
    else:
        q1 = convert_quat_wxyz_to_xyzw(q1)
        q2 = convert_quat_wxyz_to_xyzw(q2)
        orig_quats = R.from_quat(np.concatenate([q1, q2], axis=axis))
        slerp = Slerp(orig_points, orig_quats)
        new_quats = slerp(target_points).as_quat()
        return convert_quat_xyzw_to_wxyz(new_quats)


def convert_rotmat_to_quat(rotmat, quat_scalar_last=False):
    """Convert rotation matrix to quaternion."""
    if quat_scalar_last:
        return R.from_matrix(rotmat).as_quat()
    else:
        return convert_quat_xyzw_to_wxyz(
                R.from_matrix(rotmat).as_quat())


def convert_quat_to_rotmat(quat, quat_scalar_last=False):
    """Convert quaternion rotation to rotation matrix."""
    if quat_scalar_last:
        return R.from_quat(quat).as_matrix()
    else:
        return R.from_quat(
                convert_quat_wxyz_to_xyzw(quat)
        ).as_matrix()


def convert_euler_to_quat(euler, seq="xyz", quat_scalar_last=False):
    """Convert euler angles to quaternion."""
    if quat_scalar_last:
        return R.from_euler(seq, euler).as_quat()
    else:
        return convert_quat_xyzw_to_wxyz(
                R.from_euler(seq, euler).as_quat())


def convert_quat_to_euler(quat, seq="xyz", degrees=False, quat_scalar_last=False):
    """Convert quaternion rotation to euler angles."""
    if quat_scalar_last:
        return R.from_quat(quat).as_euler(seq, degrees=degrees)
    else:
        return R.from_quat(
                convert_quat_wxyz_to_xyzw(quat)
        ).as_euler(seq, degrees=degrees)


def convert_quat_to_euler_continuous(quat, seq="xyz", degrees=False,
                                     quat_scalar_last=False):
    """Convert quaternion rotation to euler removing discontinuities."""
    if not quat_scalar_last:
        quat = convert_quat_wxyz_to_xyzw(quat)

    oshape = tuple([*quat.shape[:-1], 3])
    euler_ori = R.from_quat(quat.reshape(-1, 4)).as_euler(seq).reshape(*oshape)

    euler_ori = np.unwrap(euler_ori, axis=0)    # remove discontinuities

    if degrees:
        euler_ori = np.rad2deg(euler_ori)
    return euler_ori


def qcip(quat1, quat2, keepdims=True):
    """
    Cosine of inner quaternion products [0, pi/2].

    paper (Φ3): https://www.cs.cmu.edu/~cga/dynopt/readings/Rmetric.pdf
    Based on: https://github.com/Mayitzin/ahrs/blob/master/ahrs/utils/metrics.py#L308
    """
    quat1 = quat1 / np.linalg.norm(quat1, axis=-1, keepdims=True)
    quat2 = quat2 / np.linalg.norm(quat2, axis=-1, keepdims=True)
    d = np.abs(np.sum(quat1 * quat2, axis=-1, keepdims=keepdims))
    return np.arccos(d)


def qad(quat1, quat2, eps=1e-6, keepdims=True):
    """Returns the angular distance between 2 quaternions [0, pi].

    paper (Φ6): https://www.cs.cmu.edu/~cga/dynopt/readings/Rmetric.pdf
    Based on: https://github.com/Mayitzin/ahrs/blob/master/ahrs/utils/metrics.py#L341
    """
    quat1 = quat1 / np.linalg.norm(quat1, axis=-1, keepdims=True)
    quat2 = quat2 / np.linalg.norm(quat2, axis=-1, keepdims=True)
    d = np.clip(
            2.0 * np.sum(quat1 * quat2, axis=-1, keepdims=keepdims) ** 2 - 1.0,
            -1 + eps, 1 - eps)
    return np.arccos(d)


def wrap_angle(angles, degrees=False):
    """
    Wraps angles between [-pi, pi].
    """
    ang = 180 if degrees else np.pi
    return ((angles + ang) % (2.0 * ang)) - ang
